var mediocreValidationCount = 0;

/*
 * Objects with a strong consistency can only be distributed directly to a
 * user and can not be passed indirectly. The strong consistency object has
 * a specific method that invalidates the object. This invalidation can only
 * be done after succeeding the controlFunction
 */

var candyCount = {"Twix":5};
var twixCoupon = AmbientJS.createStrongConsistencyObject({
	"uniqueName" : "twix_0001",
	"getDescription" : function() {
			return "-50% when buying a Twix";
	},
	"getStockCount" : function() {
			return candyCount["Twix"];
	},
	"redeem" : function(amount) {
			candyCount["Twix"] = candyCount["Twix"]-amount;
	}},
	"redeem",
	function (callback) {
		var client = Ti.Network.createHTTPClient({
			onload : function(e){
				if(this.responseText == "validated")
					callback();
			},
			onerror : function(e){},
			timeout : 5000
		});		
		client.open("GET", "http://192.168.1.5:3000/validatePass/" + pass.uniqueName);
		client.send();
	}
);
	
	
	var url = "http://www.appcelerator.com";
	 var client = Ti.Network.createHTTPClient({
	     // function called when the response data is available
	     onload : function(e) {
	         Ti.API.info("Received text: " + this.responseText);
	         alert('success');
	     },
	     // function called when an error occurs, including a timeout
	     onerror : function(e) {
	         Ti.API.debug(e.error);
	         alert('error');
	     },
	     timeout : 5000  // in milliseconds
	 });
	 // Prepare the connection.
	 client.open("GET", url);
	 // Send the request.
	 client.send();
	
	
function createStrongConsistencyObject(object, destroyMethod, controlFunction) {
	var referenceProxy = Ambient.createReferenceProxy(function(delegate) {
		function createLimitProxy() {
			return Ambient.createReferenceProxy(function(delegate) {
				this.onReceive = function(msg) {};
				this.onPassReference = function(reference) {
					//console.log("[WARNING] A strong consistency object can not be passed to a third party.");
				};
			});
		}
		this.destroyed = false;
		this.onReceive = function(message) {
			if ((msg.method == destroyMethod) && (!this.destroyed) && controlFunction(object)) {
				this.destroyed = true;
				delegate.onReceive(msg);
			};
		};
		this.onPassReference = function(reference){
			reference.setProxy(createLimitProxy());			
			delegate.onPassReference(reference);
		};
	});	
}



/*
 * Objects with a mediocre consistency can only be distributed directly to a
 * user and can not be passed indirectly. The mediocre consistency object has
 * a specific method that invalidates the object. This invalidation can be done
 * only once without having to succeed the controlFunction.
 */
function createMediocreConsistencyObject(object, destroyMethod, controlFunction) {
	var referenceProxy = Ambient.createReferenceProxy(function(delegate) {
		function createLimitProxy() {
			return Ambient.createReferenceProxy(function(delegate) {
				this.onReceive = function(msg) {};
				this.onPassReference = function(reference) {
					//console.log("[WARNING] A mediocre consistency object can not be passed to a third party.");
				};
			});
		}
		this.destroyed = false;
		this.onReceive = function(message) {
			if ((msg.method == destroyMethod) && (!this.destroyed) && 
				((mediocreValidationCount < 1) || controlFunction(object))) {
				this.destroyed = true;
				delegate.onReceive(msg);
				if (!controlFunction(object))
					mediocreValidationCount += 1;
				else
					mediocreValidationCount = 0;
			};
		};
		this.onPassReference = function(reference){
			reference.setProxy(createLimitProxy());			
			delegate.onPassReference(reference);
		};
	});	
}

/*
 * Objects with a weak consistency can be distributed to digital wallets of
 * other users. The weak consistency object can use a specific method that
 * can invalidate the object (e.g. redeem), without the need for additional
 * checking.
 */
function createWeakConsistencyObject(object, destroyMethod) {
	var referenceProxy = Ambient.createReferenceProxy(function(delegate) {
		function createLimitProxy(destroyedVal) {
			var callee = arguments.callee.toString();
			return Ambient.createReferenceProxy(function(delegate) {
				this.destroyed = destroyedVal;
				this.createLimitProxy = new Function("return " + callee)();
				this.onReceive = function(msg) {
					if ((msg.method == destroyMethod) && (!this.destroyed)) {
						this.destroyed = true;
						delegate.onReceive(msg);
					};
				};
				this.onPassReference = function(reference) {
					reference.setProxy(this.createLimitProxy(this.destroyed));
					delegate.onPassReference(reference);
				};
			});
		}
		this.onReceive = function(message) {
			delegate.onReceive(message);
		};
		this.onPassReference = function(reference){
			reference.setProxy(createLimitProxy(false));			
			delegate.onPassReference(reference);
		};
	});
}

function singleUseConsistencyRef(){
	var referenceProxy = AmbientJS.createReferenceProxy(function(delegate){
		function createLimitProxy() {
			var callee = arguments.callee.toString();
			return AmbientJS.createReferenceProxy(function(delegate) {
				this.used = false;
				this.onReceive = function(msg){
					if(delegate.checkConsistency(msg)) {
					    delegate.onReceive(msg);
					    this.used = false;
					} else if (!this.used) {
						delegate.onReceive(msg);
						this.used = true;
					}
				}
				this.onPassReference = function(reference) {
					reference.setProxy(this.createLimitProxy());
					delegate.onPassReference(reference);
				}
			});
		}
		this.onReceive = function(message) {
			delegate.onReceive(message);
		}
		this.onPassReference = function(reference) {
			reference.setProxy(createLimitProxy());
			delegate.onPassReference(reference);
		}
	});
	return referenceProxy;
}


function createConsistencyObject(object, consistencyLevel) {
	var refProxy;
	switch(consistencyLevel) {
		case "strong" : refProxy=strongConsistencyRef(); break;
		case "single use" : refProxy=singleUseConsistencyRef(); break;
		case "weak": refProxy = weakConstitencyRef(); break;
	} 
	AmbientJS.createObject(object, refProxy);
}